'''
Created on Nov 19, 2015

@author: lindauer
'''

import sys

if __name__ == '__main__':

    if sys.argv[1] == "1":
        print("Result of this algorithm run: UNSAT,1,1,1,12354")

    if sys.argv[1] == "2":
        print("Result of this algorithm run: UNSAT,2,3,4,12354,additional info")
